@@ -148,6 +148,7 @@
 #define DEBUG_READ 0
 #define DEBUG_WRITE 0
 
+typedef char * pchar;
 
 
 /* **************** IO ********************* */
@@ -164,7 +165,7 @@ typedef struct {
   unsigned long long fsize;
   unsigned long long bstart;
   unsigned long long bsize;
-  signed char * buffer;
+  char * buffer;
   unsigned long long dirty;
 } BIO;
 
@@ -172,7 +173,7 @@ static int read_buf(DOODLE_Logger log,
 		    void * context,
 		    int fd,
 		    unsigned long long off,
-		    signed char * buf,
+		    char * buf,
 		    unsigned long long cnt) {
   int ret;
   if (off != lseek(fd, off, SEEK_SET)) {
@@ -203,7 +204,7 @@ static void write_buf(DOODLE_Logger log,
 		      void * context,
 		      int fd,
 		      unsigned long long off,
-		      const signed char * buf,
+		      const void * buf,
 		      unsigned long long cnt) {
   int ret;
   if (off != lseek(fd, off, SEEK_SET)) {
@@ -600,9 +601,9 @@ static void WRITEULONG(BIO * fd,
   WRITEALL(fd, &v[0], (int) c);
 }
 
-static signed char * readZT(BIO * fd) {
+static char * readZT(BIO * fd) {
   unsigned int len;
-  signed char * buf;
+  char * buf;
 
   if (-1 == READUINT(fd, &len))
     return NULL;
@@ -616,7 +617,7 @@ static signed char * readZT(BIO * fd) {
 }
 
 static void writeZT(BIO * fd,
-		    signed char * buf) {
+		    const char * buf) {
   unsigned int len;
 
   len = strlen(buf);
@@ -624,13 +625,13 @@ static void writeZT(BIO * fd,
   WRITEALL(fd, buf, strlen(buf));
 }
 
-static signed char * readFN(BIO * fd,
-		     signed char ** pathTab,
+static char * readFN(BIO * fd,
+		     const pchar * pathTab,
 		     const unsigned int ptc) {
   unsigned int pid;
   unsigned int fnl;
   int slen;
-  signed char * buf;
+  char * buf;
 
   if (-1 == READUINT(fd, &pid))
     return NULL;
@@ -660,7 +661,7 @@ static signed char * readFN(BIO * fd,
 }
 
 static void writeFN(BIO * fd,
-		    signed char ** pathTab,
+		    const pchar* pathTab,
 		    const unsigned int ptc,
 		    const char * fn) {
   int i;
@@ -715,7 +716,7 @@ typedef struct DOODLE_Node {
   struct DOODLE_Node * parent;
   /* character at this position in the tree,
      pointer into tree->cis. */
-  unsigned char * c;
+  char * c;
   /* list of indices into tree->filenames */
   unsigned int * matches;
   /* how many files match here? */
@@ -746,7 +747,7 @@ typedef struct DOODLE_SuffixTree {
   /* context */
   void * context;
   /* name of the database file */
-  signed char * database;
+  char * database;
   /* file handle for database file */
   BIO * fd;
   /* size of filenames array */
@@ -758,7 +759,7 @@ typedef struct DOODLE_SuffixTree {
   /* root of the suffix tree, maybe null! */
   STNode * root;
   /* the keyword/character index string */
-  unsigned char ** cis;
+  pchar* cis;
   /* how long is cis? */
   unsigned int cisPos;
   /* how much space do we have in cis? */
@@ -789,7 +790,7 @@ const DOODLE_FileInfo * DOODLE_getFileAt(const struct DOODLE_SuffixTree * tree,
   return &tree->filenames[index];
 }
 
-static unsigned char CIS[] = {
+static char CIS[] = {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
   10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
   20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
@@ -894,7 +895,7 @@ static void processShrink(SuffixTree * tree,
 			  int ktC,
 			  int ktP,
 			  STNode * pos,
-			  int * kept) {
+			  unsigned int * kept) {
   int mark;
   STNode * next;
 
@@ -1116,7 +1117,7 @@ static STNode * lazyReadNode(SuffixTree * tree,
     }
 
     if (ret[mls].clength == 0) {
-      unsigned char c;
+      char c;
 
       if (mls == 0) {
 	if (-1 == READALL(tree->fd, &c, sizeof(unsigned char)))
@@ -1516,14 +1517,14 @@ static unsigned long long writeNode(BIO * fd,
  * time introducing the 'mls' node groups (which has the potential
  * to significantly improve performance).
  */
-static signed char * MAGIC = "DOO\0000006";
+static char * MAGIC = "DOO\0000006";
 
 /**
  * Magic string to indicate an temporary doodle database that
  * could not be completely created (the indexing/building process
  * was aborted).
  */
-static signed char * TRAGIC = "XOO\0000001";
+static char * TRAGIC = "XOO\0000001";
 
 /**
  * Create a suffix-tree (and store in file named database).
@@ -1538,7 +1539,7 @@ SuffixTree * DOODLE_tree_create_internal(DOODLE_Logger log,
   struct stat buf;
   int i;
   unsigned long long off;
-  signed char ** pathTab;
+  pchar* pathTab;
   unsigned int ptc;
   signed char magic[8];
 
@@ -1633,7 +1634,7 @@ SuffixTree * DOODLE_tree_create_internal(DOODLE_Logger log,
       return NULL;
     }
     if (ptc != 0) {
-      pathTab = MALLOC(ptc * sizeof(signed char*));
+      pathTab = MALLOC(ptc * sizeof(pchar));
       for (i=ptc-1;i>=0;i--) {
 	pathTab[i] = readZT(fd);
 	if (pathTab[i] == NULL) {
@@ -1876,7 +1877,7 @@ void DOODLE_tree_destroy(SuffixTree * tree) {
   int j;
   unsigned long long off;
   off_t wpos;
-  signed char ** pathTab;
+  pchar * pathTab;
   unsigned int ptc;
   STNode * tmp;
 
@@ -1886,7 +1887,7 @@ void DOODLE_tree_destroy(SuffixTree * tree) {
 	 ( (tree->root != NULL) &&
 	   (tree->root->modified != 0) ) ) ) {
     int fdt;
-    signed char * tdatabase;
+    char * tdatabase;
 
     tree->force_dump = 1; /* force re-dump everything! */
     tdatabase = MALLOC(strlen(tree->database) + 2);
@@ -1926,7 +1927,7 @@ void DOODLE_tree_destroy(SuffixTree * tree) {
     ptc = 0;
     pathTab = NULL;
     for (i=tree->fnc-1;i>=0;i--) {
-      signed char * fn;
+      char * fn;
       int slen;
       int xslen;
 
@@ -2164,9 +2165,9 @@ static void tree_split(SuffixTree * tree,
 }
 
 static STNode * tree_search_internal(SuffixTree * tree,
-				     const unsigned char * substring) {
+				     const char * substring) {
   STNode * pos;
-  const unsigned char * ss;
+  const char * ss;
   int i;
 
   CHECK(tree);
@@ -2245,13 +2246,13 @@ static STNode * tree_search_internal(SuffixTree * tree,
  * @return 0 on success, 1 on error
  */
 int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
-		       const unsigned char * searchString,
+		       const char * searchString,
 		       const char * fileName) {
   STNode * pos;
   STNode * spos;
-  unsigned char * cisp;
-  const unsigned char * cisp0;
-  signed char * sharedName;
+  char * cisp;
+  const char * cisp0;
+  char * sharedName;
   int i;
   int cix;
   struct stat sbuf;
@@ -2524,8 +2525,17 @@ int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
 	    continue;
 	  }
 	}
+#if ASSERTS
+	if (pos->link == NULL) {
+	  tree->log(tree->context,
+		    DOODLE_LOG_CRITICAL,
+		    _("Assertion failed at %s:%d!\n"),
+		    __FILE__, __LINE__);
+	  return 1;
+	}
+#endif
 	if (pos->link->c[0] > cisp0[0]) {
-	  if (pos->mls_size == cisp0[0] - pos->c[0]) {
+	  if (pos->mls_size == cisp0[0] - pos->c[0]) { /* expand mls range! */
 	    STNode * mlsroot;
 	    STNode * mlsnew;
 	    int mls;
@@ -2545,13 +2555,23 @@ int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
 	      mlsroot = mlsroot->parent;
 	
 	    if (pos->link->c[0] == cisp0[0] + 1) {
-	      /* JOIN two mls segments! */
+	      /* JOIN two mls segments (the new character fills the gap)! */
 	      if (pos->link->clength != 1) {
 		/* need to split tree first to make mls possible */
 		tree_split(tree,
 			   pos->link,
 			   1);
 	      }
+#if ASSERTS
+	      if ( (pos->mls_size != 1) ||
+		   (pos->link->parent != pos) ) {
+		tree->log(tree->context,
+			  DOODLE_LOG_CRITICAL,
+			  _("Assertion failed at %s:%d!\n"),
+			  __FILE__, __LINE__);
+		return 1;
+	      }
+#endif
 	      mlsnew = MALLOC(sizeof(STNode) * (mlsroot->mls_size + pos->link->mls_size + 1));
 	
 	      memcpy(mlsnew,
@@ -2580,7 +2600,7 @@ int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
 		pos->link[pos->link->mls_size-1].link->parent = &mlsnew[mlsnew->mls_size-1];
 	
 	      /* update data in new entry */
-	      mlsnew[mlsroot->mls_size].c = &CIS[mlsnew[mlsroot->mls_size-1].c[0]+1];
+	      mlsnew[mlsroot->mls_size].c = 1 + &CIS[(unsigned char) mlsnew[mlsroot->mls_size-1].c[0]];
 	
 	      /* update link from parent */
 	      if (mlsroot->parent != NULL) {
@@ -2617,8 +2637,9 @@ int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
 	      /* update pos to point to new location */
 	      free(pos->link);
 	      pos = &mlsnew[mlsroot->mls_size];
+	      /* end joining 2 mls segments */
 	    } else {
-	      /* EXTEND existing mls segment! */
+	      /* EXTEND existing mls segment by one entry! */
 	      mlsnew = MALLOC(sizeof(STNode) * (mlsroot->mls_size + 1));	
 	      memcpy(mlsnew,
 		     mlsroot,
@@ -2643,8 +2664,17 @@ int DOODLE_tree_expand(struct DOODLE_SuffixTree * tree,
 		pos->link->parent = &mlsnew[mlsnew->mls_size-1];
 	
 	      /* update data in new entry */
-	      mlsnew[mlsroot->mls_size].c = &CIS[mlsnew[mlsroot->mls_size-1].c[0]+1];
-	
+	      mlsnew[mlsroot->mls_size].c 
+		= 1 + &CIS[(unsigned char) mlsnew[mlsroot->mls_size-1].c[0]];
+#if ASSERTS	
+	      if (mlsnew[mlsroot->mls_size].c[0] != cisp0[0]) {
+		tree->log(tree->context,
+			  DOODLE_LOG_CRITICAL,
+			  _("Assertion failed at %s:%d!\n"),
+			  __FILE__, __LINE__);
+		return 1;
+	      }
+#endif
 	      /* update link from parent */
 	      if (mlsroot->parent != NULL) {
 		mlsnew[0].parent = mlsroot->parent;
@@ -3005,7 +3035,7 @@ void DOODLE_tree_truncate_modified(SuffixTree * tree,
 
   for (i=DOODLE_getFileCount(tree)-1;i>=0;i--) {
     struct stat sbuf;
-				char * fn;
+    char * fn;
     int keep;
 
     keep = 1;
@@ -3105,7 +3135,7 @@ static int tree_iterate_internal(int do_links,
  * @return 0 for not found, >0 number of results found
  */
 int DOODLE_tree_search(SuffixTree * tree,
-		       const unsigned char * substring,
+		       const char * substring,
 		       DOODLE_ResultCallback callback,
 		       void * arg) {
   STNode * pos;
@@ -3133,7 +3163,7 @@ static int tree_search_approx_internal(STNode * pos,
 				       const unsigned int approx,
 				       const int ignore_case,
 				       SuffixTree * tree,
-				       const unsigned char * ss,
+				       const char * ss,
 				       DOODLE_ResultCallback callback,
 				       void * arg) {
   int ret;
@@ -3261,7 +3291,7 @@ static int tree_search_approx_internal(STNode * pos,
 int DOODLE_tree_search_approx(SuffixTree * tree,
 			      const unsigned int approx,
 			      const int ignore_case,
-			      const unsigned char * ss,
+			      const char * ss,
 			      DOODLE_ResultCallback callback,
 			      void * arg) {
   return tree_search_approx_internal(tree->root,
@@ -3370,11 +3400,11 @@ typedef struct {
  */
 static void jni_logger(JNICTX * ctx,
 		       unsigned int level,
-		       const signed char * message,
+		       const char * message,
 		       ...) {
   size_t n;
   va_list ap;
-  signed char * buf;
+  char * buf;
   jstring arg;
   jmethodID log;
   jclass logger;
@@ -3411,7 +3441,7 @@ static void jni_logger(JNICTX * ctx,
 JNIEXPORT jlong JNICALL Java_org_gnunet_doodle_Doodle_open
 (JNIEnv * env, jclass cl, jstring filename, jobject logger) {
   JNICTX * ret;
-  const signed char * database;
+  const char * database;
   jboolean isCopy;
 
   database = (*env)->GetStringUTFChars(env,
@@ -3454,7 +3484,7 @@ JNIEXPORT void JNICALL Java_org_gnunet_doodle_Doodle_close
 JNIEXPORT void JNICALL Java_org_gnunet_doodle_Doodle_truncate
 (JNIEnv * env, jclass cl, jlong handle, jstring filename) {
   JNICTX * ret = (JNICTX*) (long) handle;
-  const signed char * fn;
+  const char * fn;
   jboolean isCopy;
 
   fn = (*env)->GetStringUTFChars(env,
@@ -3476,8 +3506,8 @@ JNIEXPORT jboolean JNICALL Java_org_gnunet_doodle_Doodle_expand
 (JNIEnv * env, jclass cl, jlong handle, jstring key, jstring filename) {
   JNICTX * ret = (JNICTX*) (long) handle;
   int i;
-  const signed char * fn;
-  const signed char * ky;
+  const char * fn;
+  const char * ky;
   jboolean isCopy;
 
   fn = (*env)->GetStringUTFChars(env,
@@ -3507,7 +3537,7 @@ JNIEXPORT jlong JNICALL Java_org_gnunet_doodle_Doodle_lastModified
 (JNIEnv * env, jclass cl, jlong handle, jstring filename) {
   JNICTX * ret = (JNICTX*) (long) handle;
   int i;
-  const signed char * fn;
+  const char * fn;
   const DOODLE_FileInfo * f;
   jboolean isCopy;
 
@@ -3557,7 +3587,7 @@ JNIEXPORT jint JNICALL Java_org_gnunet_doodle_Doodle_search
   JNICTX * ret = (JNICTX*) (long) handle;
   jclass vector;
   SearchClosure cls;
-  const signed char * ky;
+  const char * ky;
   int i;
   jboolean isCopy;
 
